import numpy as np
import logging
import math

#Helper function
def change3DSizeTo(data, shape):
    '''
    Resizes data to the new shape keeping data centralized.
    If data size is smaller, then egdes are cropped.
    If data is larger, data will be padded with zeros.
    '''
    #based in size() code found in
    #https://github.com/Biomedical-Imaging-Group/DeconvolutionLab2/blob/e9af0aba493ba137d70877154648e5583e376a81/src/main/java/signal/RealSignal.java#L531

    mz,my, mx = shape
    nz,ny,nx = data.shape

    psf1 = np.zeros( shape , dtype=data.dtype)
    
    vx = min(nx,mx)
    vy = min(ny,my)
    vz = min(nz, mz)
    ox = int((mx - nx) / 2)
    oy = int((my - ny) / 2)
    oz = int((mz - nz) / 2)

    (pz0, qz0) = (oz, 0) if oz>=0 else (0, -oz)
    (py0, qy0) = (oy, 0) if oy>=0 else (0, -oy)
    (px0, qx0) = (ox, 0) if ox>=0 else (0, -ox)

    psf1[ pz0:pz0+vz , py0:py0+vy , px0:px0+vx ] = data[ qz0:qz0+vz , qy0:qy0+vy ,  qx0:qx0+vx]

    return psf1


#An implementation of the circular() routine from DeconvolutionLab2 that shifts the data centre of the psf
#to the corners of the volume.
#This is important so that convolution of data with psf is will not shift features (like beads).
#It is assumeed that the PSF data has the psf-origin at the center of the data 
def circulify3D(data3D):
    #Check data is 3D
    if data3D.ndim != 3:
        raise Exception("Error, input data is not 3-dimensional")

    data0 = np.array(data3D)

    shape = data0.shape

    zdimhalf0 = int(shape[0]/2)
    zdimhalf1 = shape[0]-zdimhalf0

    ydimhalf0 = int(shape[1]/2)
    ydimhalf1 = shape[1]-ydimhalf0

    xdimhalf0 = int(shape[2]/2)
    xdimhalf1 = shape[2]-xdimhalf0

    if zdimhalf0 >=1 and ydimhalf0>=1 and xdimhalf0>=1:
        data0[zdimhalf1: , ydimhalf1:, xdimhalf1:] = data3D[:zdimhalf0 , :ydimhalf0 , :xdimhalf0 ] #cube 1
        data0[zdimhalf1: , ydimhalf1:, :xdimhalf1] = data3D[:zdimhalf0 , :ydimhalf0 , xdimhalf0: ] #cube 2
        data0[zdimhalf1: , :ydimhalf1, :xdimhalf1] = data3D[:zdimhalf0 , ydimhalf0: , xdimhalf0: ] #cube 3
        data0[zdimhalf1: , :ydimhalf1, xdimhalf1:] = data3D[:zdimhalf0 , ydimhalf0: , :xdimhalf0 ] #cube 4
        data0[:zdimhalf1 , ydimhalf1:, xdimhalf1:] = data3D[zdimhalf0: , :ydimhalf0 , :xdimhalf0 ] #cube 5
        data0[:zdimhalf1 , ydimhalf1:, :xdimhalf1] = data3D[zdimhalf0: , :ydimhalf0 , xdimhalf0: ] #cube 6
        data0[:zdimhalf1 , :ydimhalf1, :xdimhalf1] = data3D[zdimhalf0: , ydimhalf0: , xdimhalf0: ] #cube 7
        data0[:zdimhalf1 , :ydimhalf1, xdimhalf1:] = data3D[zdimhalf0: , ydimhalf0: , :xdimhalf0 ] #cube 8
    
    return data0

def convertToFloat32AndNormalise(data, normaliseType=None, bResetZero=True):
    '''
    Converts data to float32 format and optionally resets zero and normalises
    Parameters:
        data
        normalisetype: None for no normalisation, 'max' normalises to max, 'sum' to normalise to sum of whole volume
        bResetZero: flags to whether shift the minimum value to zero
    '''
    if data is None:
        return None
    
    ret= data.astype(np.float32)

    vmin= 0
    if bResetZero:
        vmin = ret.min()
    ret = ret-vmin

    if not normaliseType is None:
        normcorr = 1.0
        if normaliseType == 'max':
            #Normalise to maximum value
            normcorr= ret.max()
        else:
            #normalise to sum
            normcorr = np.sum(ret)
        #Check for zero before division
        if normcorr==0.0:
            logging.info("normcorr = 0. Normalisation will be skipped to prevent division by zero.")
        else:
            ret = ret/normcorr
    return ret

def convertToUint8AndFullRange(data):
    if data is None:
        return None
        
    res_256 = convertToFloat32AndNormalise(data, normaliseType='max', bResetZero=True)*256
    res_uint8 = res_256.astype('uint8')
    return res_uint8

def generateGaussPSF(shape):
    #shape = (32,32,32)
    z_mg = np.linspace(-shape[0],shape[0] ,shape[0], dtype=np.float32)
    y_mg = np.linspace(-shape[1],shape[1] ,shape[1], dtype=np.float32)
    x_mg = np.linspace(-shape[2],shape[2] ,shape[2], dtype=np.float32)

    z_mg, y_mg, x_mg = np.meshgrid( z_mg, y_mg, x_mg)

    sigma = 5

    data = 1/sigma/math.sqrt(2*math.pi) * np.exp( -0.5 * ( np.square(z_mg) + np.square(y_mg) + np.square(x_mg) ) / sigma/sigma )

    return data


#RedLionfish napari plugin

'''
Some help designing this collected from diffrent places

https://napari.org/plugins/stable/hook_specifications.html#gui-hooks
https://github.com/DragaDoncila/example-plugin/blob/main/example_plugin/_dock_widget.py
These examples show adding widget using decorator @magic_factory.
This decorator is poorly documented. I don't know how to use it.


'''
from magicgui import magic_factory
from napari_plugin_engine import napari_hook_implementation

from napari.types import LabelsData, ImageData, LayerDataTuple
from napari.layers import Image

import RedLionfishDeconv as rl

import logging
logging.basicConfig(level=logging.INFO)

''' The parameters after the decorator setup the title and other properties of the widget window.

The parameters in the FUNCTION will set the elements that will be shown, but things like comboboxes
or sliders will have no limits set, this should be done in the decorator.

see also
https://napari.org/guides/stable/magicgui.html

Note that the @magic_factory and @magicgui behave in similar way.
'''


# @magic_factory (
#     call_button="Test button0" #,
#     #iterations={'max':100}
#     )
# def test_widget0(
#     data: "ImageData", #Input is data that can be selected
#     psfdata:"ImageData",
#     iterations=10 #User chooses number of RL iterations
#     #useGPU = True #
#     ) -> LabelsData: #Result is data
    
#     print(f"you have selected {data}")


# @magic_factory (
#     call_button="Test button1" ,
#     iterations={'max':100}
#     )
# def test_widget1(
#     data: ImageData, #Input is data that can be selected
#     psfdata:ImageData,
#     iterations=10, #User chooses number of RL iterations
#     useGPU = True 
#     ) -> LabelsData: #Result is data

#     print("You clicked the button.")
#     print(f"iterations = {iterations}")
#     if not data is None:
#         print(f"data shape = {data.shape}")
#         print(f"type(data) = {type(data)}")
#         print(f"data.dtype = {data.dtype}")

#     if not psfdata is None:
#         print(f"psfdata shape = {psfdata.shape}")
#         print(f"type(psfdata) = {type(psfdata)}")
#         print(f"psfdata.dtype = {psfdata.dtype}")


# #Really basic
# @magic_factory (call_button="Test button2")
# def test_widget2(img_layer: Image):
#     print(f"you have selected {img_layer}")


#The widget
@magic_factory (
    call_button="Go" ,
    iterations={'max':100} 
    )
def RedLionfish_widget(
    data: ImageData, #Input is data that can be selected
    psfdata:ImageData,
    iterations=10, #User chooses number of RL iterations
    useGPU = True 
    ) -> LayerDataTuple: #Result is a LayerDataTuple, like (data, {dict_properties})

    print(f"iterations = {iterations}")

    ret = None

    datares_uint8=None
    if not data is None and not psfdata is None:
        #Print information before calculation
        print(f"data.shape = {data.shape} , type(data) = {type(data)} , data.dtype = {data.dtype}")
        print(f"psfdata.shape = {psfdata.shape} , type(psfdata) = {type(psfdata)} , psfdata.dtype = {psfdata.dtype}")
        
        if data.ndim==3 and psfdata.ndim==3: #For now, only 3D is supported
            method = 'cpu' #Default
            if useGPU:
                method='gpu'
            #Run the deconvolution
            datares_uint8 = rl.doRLDeconvolutionFromNpArrays(data, psfdata, niter= iterations, method = method ,resAsUint8=True )

            ret = ( datares_uint8 , { 'name':'RL-deconvolution'})
        else:
            print("Data or PSF is not 3-dimensional.")

    return ret



@napari_hook_implementation
def napari_experimental_provide_dock_widget():
    # you can return either a single widget, or a sequence of widgets
    #return [test_widget0, test_widget1]
    #return [RedLionfish_widget,test_widget2, test_widget1]
    return RedLionfish_widget


#Several versions of Richardson-Lucy deconvolution using scipy.signal convolution

import numpy as np
import scipy.fft

from .helperfunctions import *
import logging

def doRLDeconvolution12(data_np , psf_np , *, niter=10, callbkTickFunc=None):
    #RL deconvolution based on doRLDeconvolution11()
    # Reverse engineered parts of https://scikit-image.org/docs/dev/api/skimage.restoration.html#skimage.restoration.richardson_lucy
    # and also scipy.signal convolution fft code for faster processing
    #https://github.com/scipy/scipy/blob/v1.7.1/scipy/signal/signaltools.py#L1293-L1413
    #for mode='same', method='fft', fftconvolution() #
    #There is a small difference here compared to scikit-image version in the sense that the psf is normalised to the sum.
    #This version tries to do all the work with arrays with preferred size of next_fast_len
    #Only at the end of the calculation it gets the shape needed

    #Convert and normalise
    data_np_norm = convertToFloat32AndNormalise(data_np)
    psf_np_norm = convertToFloat32AndNormalise(psf_np , normaliseType='sum', bResetZero=False) #Normalise to sum

    s1 = data_np_norm.shape
    s2 = psf_np_norm.shape

    shape = [(s1[i] + s2[i] - 1) for i in range(data_np_norm.ndim)]
    fshape = [scipy.fft.next_fast_len(shape[a], True) for a in range(len(shape))]

    #Resize both the data and psf to the preferred fshape
    data_shaped = change3DSizeTo(data_np_norm, fshape)
    psf_shaped = change3DSizeTo(psf_np_norm, fshape)

    #Circulify psf
    psf1 = circulify3D(psf_shaped)

    #Precalculated psf_fft
    psf_fft= scipy.fft.rfftn(psf1)
    psf_flip = np.flip(psf1)
    psf_flip_fft = scipy.fft.rfftn(psf_flip) #FFT of the flipped psf

    xn1 = np.array(data_shaped) #initialize copy
    for i in range(niter):
        xn=xn1
        xn_fft = scipy.fft.rfftn(xn) 
        U0 = np.multiply(xn_fft ,psf_fft)
        u0 = scipy.fft.irfftn(U0)

        u0 = np.where(u0==0, 1e-14, u0) #Fix zero values giving a finite value
            
        p = np.divide(data_shaped,u0)

        U1 = scipy.fft.rfftn(p)
        U2 = np.multiply(U1 , psf_flip_fft)
        u2 = scipy.fft.irfftn(U2)

        xn1 = np.multiply(xn, u2)

        if not callbkTickFunc is None:
            callbkTickFunc()
    
    #Get the central part of the result
    #xn1_central = _centered(xn1, s1)
    xn1_central = change3DSizeTo(xn1, s1) #To original shape

    # data_deconv_norm_256 = convertAndNormalise(xn1_central)*256
    # data_deconv_uint8 = data_deconv_norm_256.astype('uint8')
    
    data_deconv_uint8 = convertToUint8AndFullRange(xn1_central)
    return data_deconv_uint8


#default
def doRLDeconvolution_DL2_4(data_np , psf_np ,*, niter=10, callbkTickFunc=None):
    #RL deconvolution based in DeconvolutionLab2 with optional parameter for normalising inputs
    #Mimics DeconvolutionLab2 (DL2) as best as possible
    #https://github.com/Biomedical-Imaging-Group/DeconvolutionLab2/blob/master/src/main/java/deconvolution/algorithm/RichardsonLucy.java
    #However it prepares the psf to have the same size as the input data
    #In DeconvolutionLab2 only psf is normalised to psf sum
    #
    #This version uses scipy
    #It is modified from the DeconvolutionLab2 algorithm
    # the fft of the psf is not conjugated, but instead is flipped before conjugation
    # The developers probably thought that the 'reality condition' applies
    # FT[f](-y) = FT[f]*
    # but that is not true, as code tests proved (tests.ipynb)
    # In this version, as in the original formula for RL, the flipped psf is used for the second convolution
    # (see tests.ipynb)
    
    logging.info("doRLDeconvolution_DL2_4() (CPU)")
    #Convert and normalise
    data_np_norm =convertToFloat32AndNormalise(data_np) #don't normalise

    #Check last axis is even size, otherwise it will give error
    if data_np_norm.shape[-1] %2 != 0: #odd number
        logging.info("Data last axis size is an odd number. Padding with zeros to make size even to prevent errors")
        s = data_np_norm.shape
        shape_fix = (s[0], s[1], s[2]+1)
        data_fix = np.empty(shape_fix, dtype = data_np_norm.dtype)
        data_fix[:,:,:-1] = data_np_norm[:,:,:]
        data_fix[:,:,-1] = 0
        data_np_norm = data_fix

    # psf_norm = psf0/sum#
    psf_norm = convertToFloat32AndNormalise(psf_np, normaliseType='sum',bResetZero=False)

    psf0 = change3DSizeTo(psf_norm, data_np_norm.shape)
    #psf1 should have the same dimensions as data

    #Do circulirisation of psf data before starting the RL algorithm
    psf1 = circulify3D(psf0)

    #Precalculated psf_fft
    psf_fft= scipy.fft.rfftn(psf1)
    psf_flip = np.flip(psf1)
    psf_flip_fft = scipy.fft.rfftn(psf_flip) #FFT of the flipped psf

    xn1 = np.array(data_np_norm) #initialize copy
    for i in range(niter):
        xn=xn1
        xn_fft = scipy.fft.rfftn(xn)  
        U0 = np.multiply(xn_fft ,psf_fft)
        u0 = scipy.fft.irfftn(U0)

        u0 = np.where(u0==0, 1e-14, u0) #Fix zero values giving a finite value

        p = np.divide(data_np_norm,u0)

        U1 = scipy.fft.rfftn(p)
        U2 = np.multiply(U1 , psf_flip_fft)
        u2 = scipy.fft.irfftn(U2)

        xn1 = np.multiply(xn, u2)

        if not callbkTickFunc is None:
            callbkTickFunc()

    data_deconv_uint8 = convertToUint8AndFullRange(xn1)
    return data_deconv_uint8